/**** SimCore since 2002-05-15 by Kenji KISE ****/ 
/************************************************/
#include "scalar.h"

/** Data Cache Unit                            **/
/************************************************/
Dcache::Dcache(Env_s *env){
  ev = env;
  nw = DCLINE_SIZE/DDATA_SIZE;
  size = ev->sc2->dcache_entry;
  buf  = new dcache_line[size];

  stall=0;
  for(int i=0; i<size; i++){
    buf[i].valid = 0;
    buf[i].dirty = 0;
    buf[i].tag   = 0;
  }
}

/************************************************/
void Dcache::Init(chip_wire *w){}

/************************************************/
void Dcache::flush(){
  stall=0;
  for(int i=0; i<size; i++){
    if(buf[i].valid && buf[i].dirty){
      st_line(buf[i].tag << DCLINE_SFT);
    }
    buf[i].valid = 0;
    buf[i].dirty = 0;
    buf[i].tag   = 0;
  }
}

/************************************************/
void Dcache::st_line(ADDR_TYPE adr){
  int       index = (adr>>DCLINE_SFT) % size;
  dcache_line *cl = &buf[index];
  ADDR_TYPE tag   = buf[index].tag;

  if(cl->valid && cl->dirty){
    for(int i=0; i<nw; i++){
      data_t a;
      data_t *d = &cl->d[i];
      a = ((tag<<DCLINE_SFT) + DDATA_SIZE*i);
      ev->mem->st_nbyte(DDATA_SIZE, &a, d);
    }
  }
}

/************************************************/
void Dcache::ld_line(ADDR_TYPE adr){
  ADDR_TYPE tag    = (adr>>DCLINE_SFT);
  int       index  = (adr>>DCLINE_SFT) % size;
  dcache_line *cl = &buf[index];

  cl->valid = 1;
  cl->dirty = 0;
  cl->tag   = tag;
  for(int i=0; i<nw; i++){
    data_t a;
    data_t *d = &cl->d[i];
    a = ((tag<<DCLINE_SFT) + DDATA_SIZE*i);
    ev->mem->ld_nbyte(DDATA_SIZE, &a, d);
  }
}

/************************************************/
void Dcache::Step(chip_wire *w){
  w->dcache_stall = stall; /* default */
  
  if(w->dc_radr==0) return;

  /***** Waiting for the data *****/
  if(stall){
    w->dcache_stall = stall;
    return;
  }

  /** Load Instruction **/
  ADDR_TYPE adr = w->dc_radr;
  ADDR_TYPE tag = (adr>>DCLINE_SFT);
  int     index = (adr>>DCLINE_SFT) % size;
  int    offset = (adr>>DDATA_SFT ) % nw;
  dcache_line *cl = &buf[index];
  
  /**** D-Cache Hit ****/
  if(cl->valid && cl->tag==tag){ 
    stall = 0;
    w->dcache_stall = 0;
    ev->e2->dcache_access++;
    data_t tmp = cl->d[offset];

    /*** Store -> Load data forwarding ***/
    if((w->dc_radr>>3)==
       (w->dc_wadr>>3)){
      data_t dat;
      DATA_TYPE mask;
      int ci = get_code(w->dc_wir);
      set_st_data(ci, st_bytes(ci),
		  &w->dc_wadr,
		  &w->dc_wdat, &dat, &mask);
      tmp= ((cl->d[offset] & mask) | dat);
    }
    
    {
      int ci = get_code(w->dc_rir);
      set_ld_data(ci, ld_bytes(ci),  &w->dc_radr, 
		  &tmp, &w->dc_rdat);
    }
  }
  /**** D-Cache Miss ****/
  else{ 
    stall = ev->sc2->dcache_penalty;
    mis_adr = adr;
    w->dcache_stall = stall;
    st_line(adr);
    ld_line(adr);
    ev->e2->dcache_mis++;
  }
}

/************************************************/
void Dcache::Update(chip_wire *w){
  if(stall>0) stall--;

  if(w->dc_wadr==0) return;

  data_t dat;
  DATA_TYPE mask;
  int ci = get_code(w->stg[5]->ir);
  set_st_data(ci, st_bytes(ci), &w->dc_wadr,
	      &w->dc_wdat, &dat, &mask);
  
  /***** Store to the cache *****/
  ADDR_TYPE adr = w->dc_wadr;
  ADDR_TYPE tag = (adr>>DCLINE_SFT);
  int     index = (adr>>DCLINE_SFT) % size;
  int    offset = (adr>>DDATA_SFT ) % 
    (DCLINE_SIZE/DDATA_SIZE);
  dcache_line *cl = &buf[index];
  
  /**** D-Cache Hit ****/
  if(cl->valid && cl->tag==tag){ 
    cl->d[offset] = ((cl->d[offset] 
		      & mask) | dat);
    cl->dirty=1;
  }
  /**** D-Cache Miss ****/
  else{  /** Store to the memory **/
    ev->mem->st_8byte(&w->dc_wadr, &dat, mask);
  }
}

/** Instruction Cache Unit                     **/
/************************************************/
Icache::Icache(Env_s *env){
  stall = 0;
  ev = env;
  nw = ICLINE_SIZE/ICODE_SIZE;
  size = ev->sc2->icache_entry;
  buf  = new icache_line[size];
  for(int i=0; i<size; i++){
    buf[i].valid = 0;
    buf[i].tag   = 0;
  }
}

/************************************************/
void Icache::Init(chip_wire *w){}

/************************************************/
void Icache::Update(chip_wire *w){
  if(stall>0) stall--;
  if(stall==1) ld_line(mis_pc);
}

/************************************************/
void Icache::ld_line(ADDR_TYPE pc){
  ADDR_TYPE tag    = (pc>>ICLINE_SFT);
  int       index  = (pc>>ICLINE_SFT) % size;
  icache_line *e = &buf[index];

  e->tag   = tag;
  e->valid = 1;
  ADDR_TYPE adr = tag << ICLINE_SFT;
  for(int i=0; i<nw; i++){
    data_t a;
    a = (adr + ICODE_SIZE*i);
    ev->mem->ld_inst(&a, &e->d[i]);
  }
}

/************************************************/
void Icache::Step(chip_wire *w){
  ADDR_TYPE pc     = w->pcgen_pc;
  ADDR_TYPE tag    = (pc>>ICLINE_SFT);
  int       index  = (pc>>ICLINE_SFT) % size;
  int       offset = (pc>>ICODE_SFT) % nw;
  
  icache_line *e = &buf[index];
  
  if(stall){ /** Waiting for the data **/
    w->icache_stall = stall;
  }
  else{
    /**** I-Cache Hit ****/
    if(e->valid && e->tag==tag){ 
      stall           = 0;
      w->icache_stall = 0;
      ev->e2->icache_access++;
    }
    /**** I-Cache Miss ****/
    else{ 
      stall           = ev->sc2->icache_penalty;
      w->icache_stall = ev->sc2->icache_penalty;
      mis_pc = pc;
      ev->e2->icache_mis++;
    }
  }
  w->ic_out = (stall) ? 0 : e->d[offset];
}
/************************************************/
