/**** SimCore since 2002-05-15 by Kenji KISE ****/ 
/************************************************/
#include <unistd.h> 
#include "scalar.h"

/** system_config for RealScalar               **/
/************************************************/
system_config_s::system_config_s(char **opt){
  btb_entry      = 512;
  icache_entry   = 512;
  icache_penalty = 10;
  dcache_entry   = 512;
  dcache_penalty = 10;
}

/** print usage for SimCore_RealScalar         **/
/************************************************/
void usage_rs(){
  char UsageMessage[] =  "\
Usage: SimCore_RealScalar [-option] command\n\
 -d: debug mode\n\
 -z: verify mode\n\
 -a: print simple evaluation result\n\
 -c: print systemcall trace\n\
 -t[num][m]: print execution trace\n\
 -e[num][m]: end simulation at num code run. \
m means million\n\
 -f[num]: set fd_adjust to verify simulation\n\
 -v[num]: verbose mode, prints '.'\
by VERBOSE_INTERVAL\n\
 -i[file_name]: file_name as STDIN input file\n\
";
  printf("\n%s", UsageMessage); exit(0);
}

/************************************************/
void scalar_chip::print_result(){
  class evaluation_result   *e  = ev->e;
  printf("===\n");
  printf("=== BTB entry            : %11d\n", 
         ev->sc2->btb_entry);
  printf("=== Branch_Miss_Penalty  : %11d\n", 
         PBRED_MISS_PENALTY);
  printf("=== Conditional   Branch : %11lld\n",
         ev->e2->c_branch);
  
  printf("=== Cond Branch HIT      : %11lld\n",
         ev->e2->c_branch_hit);
  printf("=== Cond Branch MISS     : %11lld\n",
         ev->e2->c_branch_mis);
  printf("=== Cond Branch HIT Rate : %11.3f\n",
         (double)ev->e2->c_branch_hit / 
         (double)
	 (ev->e2->c_branch_hit + ev->e2->c_branch_mis));
  printf("===\n");
  
  printf("=== UnConditional Branch : %11lld\n",
         ev->e2->unc_branch);
  printf("===\n");
  
  uint_64t icache_hit = 
    ev->e2->icache_access - ev->e2->icache_mis;
  printf("=== Icache entry         : %11d\n",
         ev->sc2->icache_entry);
  printf("=== Icache Miss Penalty  : %11d\n",
         ev->sc2->icache_penalty);
  printf("=== Icache ACCESS        : %11lld\n",
         ev->e2->icache_access);
  printf("=== Icache HIT           : %11lld\n",
         icache_hit);
  printf("=== Icache MISS          : %11lld\n",
         ev->e2->icache_mis);
  printf("=== Icache HIT Rate      : %11.3f\n",
         (double)icache_hit / 
         (double)(ev->e2->icache_access));
  printf("===\n");

  printf("=== Dcache entry         : %11d\n",
         ev->sc2->dcache_entry);
  printf("=== Dcache Miss Penalty  : %11d\n",
         ev->sc2->dcache_penalty);

  uint_64t dcache_hit = 
    ev->e2->dcache_access - ev->e2->dcache_mis;
  printf("=== Dcache ACCESS        : %11lld\n",
         ev->e2->dcache_access);
  printf("=== Dcache HIT           : %11lld\n",
         dcache_hit);
  printf("=== Dcache MISS          : %11lld\n",
         ev->e2->dcache_mis);
  printf("=== Dcache HIT Rate      : %11.3f\n",
         (double)(dcache_hit)/ 
         (double)(ev->e2->dcache_access));
  printf("===\n");
  
  printf("=== Clock Cycles         : %11lld\n",
         ev->e2->clock);
  printf("=== CPI(cycles per inst.): %11.3f\n", 
         (double)ev->e2->clock/
	 (double)e->retired_inst);
  fflush(stdout);
}

#define FLYING_CODE_MASK 0x3ff
#define FLYING_CODE_MAX (FLYING_CODE_MASK+1)
/************************************************/
InstBuf::InstBuf(Env *ev){ /* Constructor */
  buf = new instruction_s*[FLYING_CODE_MAX];
  for(int i=0; i<FLYING_CODE_MAX; i++){
    buf[i] = new instruction_s(ev);
  }
}

/************************************************/
instruction_s *InstBuf::alloc(){
  static int head = 0;
  head++;
  head &= FLYING_CODE_MASK;
  return buf[head];
}

#define WD_INT 0xff /** watch dog interval **/
/************************************************/
void scalar_chip::house_keeping(){
  system_config       *sc  = ev->sc;
  evaluation_result   *e   = ev->e;
  evaluation_result_s *e2  = ev->e2;
  static int wd_cnt = 0; /* watch dog count */

  e2->clock++;

  /*** watch dog timer check by 0xff cycles ***/
  if((e2->clock&WD_INT)==WD_INT){
    if(wd_cnt==0){
      printf("\nWatch Dog Timer: retired %lld\n",
             e->retired_inst);
      exit(1);
    }
    wd_cnt=0;
  }

  if(sc->trace_flag!=0 &&
     sc->trace_flag<=(e->retired_inst+1))
    print_pipeline();

  instruction_s *pt = stg[5];
  /** write back stage has a code to retire **/
  if(pt!=Inull){
    e->retired_inst++;
    wd_cnt++;

    if(sc->verify_flag) verify();
    if(sc->debug_flag) ev->deb->debugmode(pt);

    if(sc->verbose_interval && 
       (e->retired_inst % 
	sc->verbose_interval)==0){
      write(1, ".", 1);
    }

    if(sc->end_tc && e->retired_inst>=sc->end_tc)
      ev->sys->running = 0;
    
    /**** branch instruction_s *****/
    if(pt->BR || pt->Op==OP_JSR){
      /*** un-conditional branch ***/
      if(pt->Op==OP_JSR || 
	 pt->CI==_BR__ || pt->CI==_BSR_){
	e2->unc_branch++;
      }
      /*** conditional branch    ***/
      else{
	e2->c_branch++;
	if(pt->Ppc==pt->Npc) e2->c_branch_hit++;
	else                 e2->c_branch_mis++;
      }
    }
  } 
}

/* print pipeline state if -t option is used.  */
/************************************************/
void scalar_chip::print_pipeline(){
  if(ev->e2->clock==1){
    printf(" cycle|  IF   ID   RR   ");
    printf("EX   MM   WB  |la:D:IC:DC:\n");
  }
  printf("%06lld| ", ev->e2->clock);
  for(int i=0; i<P_DEPTH; i++){
    if(stg[i]==Inull) printf("---- ");
    else{
      printf("%04llx ", 
             stg[i]->Cpc&0xffff);
    }
  }
  if(w->code_stall) printf("|%2d:",insn_latency);
  else              printf("|--:");

  if(w->data_stall) printf("d:");
  else              printf("-:");

  if(w->icache_stall) 
    printf("%2d:", w->icache_stall);
  else printf("--:");

  if(w->dcache_stall) 
    printf("%2d|", w->dcache_stall);
  else printf("--|");

  if(stg[5]->LD) printf("ld");
  else if(stg[5]->ST) printf("st");
  else if(stg[5]->CM) printf("cm");
  else if(stg[5]->BR) printf("br");
  else printf("  ");
  printf("|");
  
  if(stg[5]->ir){
    printf("%04lld ", ev->e->retired_inst + 1);
    int id = stg[5]->CI;
    printf("%02x.%04x", id>>16, id & 0xffff);
  }
  printf("\n");
  
  if(w->flush_bpr){
    printf("Branch_Miss_Prediction %05lld", 
           ev->e2->bpred_miss + 1);
    printf(" (%04llx -> %04llx)\n\n", 
           stg[3]->Ppc & 0xffff,
           stg[3]->Npc & 0xffff);
  }
}

/************************************************/
void scalar_chip::verify(){
  simple_chip *c = s_chip;
  ADDR_TYPE c_cpc = c->ev->as->rr(0);
  c->step();

  int error = 0;
  for(int i=1; i<=64; i++)
    if(c->ev->as->rr(i)!=ev->as->rr(i)) error=1;
  if(c_cpc!=stg[5]->Cpc) error=1;

  if(error){
    instruction   *cp = c->p;
    instruction_s *sp = stg[5];
    
    printf("** Verify Error:% lld code retire\n", 
           ev->e->retired_inst);
    printf("** simple_chip / real_scalar\n\n");
    printf("Cpc:%16llx %16llx\n", 
           c_cpc, stg[5]->Cpc);
    printf("Npc:%16llx %16llx\n", 
	   cp->Npc, sp->Npc);
    printf("Rav:%16llx %16llx\n", 
           cp->Rav, sp->Rav);
    printf("Rbv:%16llx %16llx\n", 
           cp->Rbv, sp->Rbv);
    printf("Rcv:%16llx %16llx\n", 
           cp->Rcv, sp->Rcv);
    printf("Adr:%16llx %16llx\n", 
           cp->Adr, sp->Adr);

    for(int i=1; i<=32; i++){
      print_register_name(i-1);
      printf("%02d:%016llx %016llx", i-1, 
             c->ev->as->rr(i), ev->as->rr(i));
      if(c->ev->as->rr(i)!=ev->as->rr(i)) 
        printf("*");
      else 
        printf(" ");
      printf(" f:%016llx %016llx", 
             c->ev->as->rr(i+32), 
             ev->as->rr(i+32));
      if(c->ev->as->rr(i+32)!=ev->as->rr(i+32))
        printf("*");
      else 
        printf(" ");
      printf("\n");
    }
    exit(1);
  }
}

/************************************************/
scalar_chip::scalar_chip(char *prog, char **opt){
  ev       = new Env_s();
  ev->sc   = new system_config(prog, opt);
  ev->e    = new evaluation_result();
  ev->as   = new architecture_state(ev);
  ev->mem  = new memory_system(ev);
  ev->deb  = new debug(ev);
  ev->sys  = new system_manager(ev);

  ev->sc2  = new system_config_s(opt);
  ev->e2   = new evaluation_result_s();

  pc_gen   = new PCgen(ev);
  b_pred   = new Bpred(ev);
  icache   = new Icache(ev);
  dcache   = new Dcache(ev);
  inst_buf = new InstBuf(ev);
  w        = new chip_wire();

  Inull    = new instruction_s(ev);
  Inull->ir = 0;
  Inull->Slot();                 

  stg   = new instruction_s*[P_DEPTH];
  w->stg= new instruction_s*[P_DEPTH];
  for(int i=0; i<P_DEPTH; i++){
    w->stg[i] = stg[i] = Inull;
  }
  s_chip = NULL;

  insn_latency = 1;
}

/************************************************/
scalar_chip::~scalar_chip(){   /** destructor **/
  print_evaluation_result(ev);
  if(!ev->sc->simple_result) print_result();
}

/************************************************/
void scalar_chip::Init(){
  w->pcgen_pc = pc_gen->pc;
  
  w->stg[0] = Inull;
  for(int i=1; i<P_DEPTH; i++) w->stg[i]=stg[i-1];
  
  instruction_s *p1 = w->stg[5];
  w->dc_wir  = (p1->ST) ? p1->ir  : 0;
  w->dc_wadr = (p1->ST) ? p1->Adr : 0;
  w->dc_wdat = (p1->ST) ? p1->Rav : 0;

  instruction_s *p2 = w->stg[4];
  w->dc_rir  = (p2->LD) ? p2->ir  : 0;
  w->dc_radr = (p2->LD) ? p2->Adr : 0;
  w->dc_rdat = 0;
}

/** return the instruction_s latency           **/
/** value must be modified concerning config.  **/
/************************************************/
int scalar_chip::latency(char Op, int CI){
  if(Op==OP_INTM) return 4;
  if(Op==OP_FLTI) return 5;
  return 1; /* default code latency */
}

/************************************************/
void scalar_chip::IDecode(){
  instruction_s *pt = w->stg[1];

  pt->Slot();
  pt->Issue();

  pt->Ra =(pt->Ar << 5) |  pt->RA;
  pt->Rb =(pt->Br << 5) |  pt->RB;
  pt->Rc =(pt->WB==31) ? 0xff:
    (WR(pt->Op)<< 5) | pt->WB;
  pt->Rm = (!pt->CM) ? 0xf0 : 
    (WR(pt->Op)<< 5) | pt->WB;
}

/************************************************/
void scalar_chip::RegRead(){
  w->stg[2]->RegisterRead();
}

/************************************************/
void scalar_chip::Execute(){
  static int busy=0;
  static int mis;
  data_t target_pc = 0;

  w->code_stall = 0;
  w->flush_bpr  = 0;
  w->pcgen_bpr  = 0;
    
  instruction_s *pt = w->stg[3];
  if(pt==Inull) return;

  if(!busy){ /** fire new instruction_s **/
    insn_latency = latency(pt->Op, pt->CI);
    pt->Execute();
    target_pc = pt->ev->as->pc;
    mis=(pt->Ppc!=target_pc);
    
    if(insn_latency==1){
      w->flush_bpr = (mis) ? 1 : 0;
      w->pcgen_bpr = (mis) ? target_pc : 0;
      busy=0;
    }
    else{
      busy=1;
      w->code_stall=1;
    }
  }
  else{ /** executing multi-cycle inst. **/
    if(insn_latency>1) insn_latency--;
    if(insn_latency==1){
      w->flush_bpr = (mis) ? 1 : 0;
      w->pcgen_bpr = (mis) ? target_pc : 0;
      busy=0;
    }
    else{
      w->code_stall=1;
    }
  }
}

/************************************************/
void scalar_chip::WriteBack(){
  instruction_s *p = w->stg[5];

  int br = (p->Op==OP_JSR || p->BR);
  w->bpred_wtkn = (br) & 
    (p->Cpc+4 != p->Npc);
  w->bpred_wpc  = (br) ? p->Cpc : 0;
  w->bpred_wnpc = (br) ? p->Npc : 0;
  
  /***** Data Cache Flush by PAL*****/
  if(p!=Inull && p->Op==OP_PAL) dcache->flush(); 

  if(p!=Inull){
    ev->as->pc = p->Npc;
    p->WriteBack();  /** Write Back **/
  }
  
  int pal = (p!=Inull && p->Op==OP_PAL);
  /* sigreturn updates as->ps */
  w->pcgen_pal = (pal) ? ev->as->pc : 0;
  w->flush_pal = (pal) ? 1 : 0;
}

/************************************************/
void scalar_chip::Step(){
  /***** data deps stall calculation *****/
  if(w->stg[2]==Inull || w->stg[3]==Inull){
    w->data_stall = 0;
  }
  else{
    instruction_s *pr = w->stg[2];
    instruction_s *pw = w->stg[3];

    /** dependency: RegRead after Load **/
    int w1  = pw->LD && pw->WB!=31 &&
      (pr->Ra == pw->Rc || pr->Rb == pw->Rc);
    
    /** depencency: CMOV    after Load **/
    int w2 = pr->CM && pw->LD && pw->WB!=31 &&
      pr->Rm == pw->Rc;

    w->data_stall = w1 | w2;
  }
}

/************************************************/
void scalar_chip::Update1(){
  /***** Set fetched code info. *****/
  if(!w->icache_stall){
    instruction_s *p = inst_buf->alloc();
    p->ir  = w->ic_out;
    p->Cpc = w->pcgen_pc;
    p->Npc = w->pcgen_npc;
    p->Ppc = (w->bpred_tkn) ? w->bpred_targ :
      w->pcgen_npc;
    w->stg[0] = p;
  }

  /***** Set Rcv from the data cache *****/
  if(w->stg[4]->LD) w->stg[4]->Rcv=w->dc_rdat;
}

/************************************************/
void scalar_chip::Update2(){
  /**** Data forwarding to Rav & Rbv & Adr ****/
  if(w->stg[2]!=Inull){
    instruction_s *pr = w->stg[2];
    for(int i=5; i>=3; i--){
      instruction_s *pw = w->stg[i];
      if(pw!=Inull){
	if(pr->Ra==pw->Rc) pr->Rav=pw->Rcv;
	if(pr->Rb==pw->Rc) pr->Rbv=pw->Rcv;
	if(pr->Rm==pw->Rc) pr->Adr=pw->Rcv;
      }
    }
  }

}

/************************************************/
void scalar_chip::Update3(){
  int hazard = w->flush_pal | w->dcache_stall | 
    w->flush_bpr | w->code_stall | w->data_stall;
  if(!hazard){
    for(int i=0; i<P_DEPTH; i++){
      stg[i] = w->stg[i];
    }
  }
  else {
    /******** Update Pipeline Registers ********/
    int p_null   = 0x000000; /* pipe null   */
    int p_update = 0x111111; /* pipe update */

    /***** (1) Flush by PAL Code *****/
    if(w->flush_pal){
      p_null   = 0x111110;
      p_update = 0x111111;
    }
    /***** (2) Stall by D-cache Miss *****/
    else if(w->dcache_stall){
      p_null   = 0x000010;
      p_update = 0x000011;
    }
    /***** (3) Flush by Branch Mis-pred *****/
    else if(w->flush_bpr){
      p_null   = 0x111000;
      p_update = 0x111111;
      ev->e2->bpred_miss++; /** evaluation **/
    }
    /***** (4) Stall by Code Latency  *****/
    else if(w->code_stall){
      p_null   = 0x100100;
      p_update = 0x000111;
    }
    /***** (5) Stall by Data Dependene *****/
    else if(w->data_stall){
      p_null   = 0x101000;
      p_update = 0x001111;
    }

    static int msk[P_DEPTH] ={
      0x100000, 0x010000, 0x001000, 
      0x000100, 0x000010, 0x000001};

    for(int i=0; i<P_DEPTH; i++){
      if(p_null   & msk[i]) w->stg[i] = Inull;
      if(p_update & msk[i]) stg[i] = w->stg[i];
    }
  }
}

/************************************************/
void scalar_chip::main_loop(){
  while(ev->sys->running){
    /******** (1) Init    ********/
    Init();

    /******** (2) Step    ********/
    Step();
    pc_gen->Step(w); /* IF */
    b_pred->Step(w); /* IF */
    icache->Step(w); /* IF */
    dcache->Step(w); /* MM */
    IDecode();       /* ID */
    RegRead();       /* RR */
    Execute();       /* EX */
    WriteBack();     /* WB */

    /******** (3) Update  ********/
    Update1();
    Update2();
    Update3();

    pc_gen->Update(w);
    b_pred->Update(w);
    icache->Update(w);
    dcache->Update(w);
    
    /******** etc.        ********/
    house_keeping();
  }
}

/************************************************/
int main(int argc, char **argv){
  printf("%s/%s %s\n", PROJ, RS_NAME, RS_VER); 
  if(argc==1) usage_rs();

  char *p    = argv[argc-1]; /* program name */
  char **opt = argv;         /* options      */
  scalar_chip *c = new scalar_chip(p, opt);

  if(c->ev->sc->verify_flag){
    char *op[OPTION_MAX];
    option_from_file(op, "verify.txt");
    c->s_chip = new simple_chip(p, op);
  }
  
  c->main_loop();
  
  delete c;
  return 0;
}
/************************************************/
