/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include "SimCore/define.h"

#define PROJ "SimCore"
#define RS_NAME "Alpha RealScalar"
#define RS_VER  "Version 0.9.30 2003-10-03"

#define BUF_SIZE 4096
#define B_TKN 1 /* Branch Taken   */
#define B_UNT 0 /* Branch Untaken */

/* pipeline depth       */
#define P_DEPTH             6     
#define PBRED_MISS_PENALTY  2

/** instruction for RealScalar                 **/
/************************************************/
class instruction_s : public instruction{
 public:
  char Ra;      /*  6b: for dependency check   */
  char Rb;      /*  6b: for dependency check   */
  char Rm;      /*  6b: for dependency check   */
  char Rc;      /*  6b: for dependency check   */
  data_t  Ppc;  /* bpred predicted PC          */
  instruction_s(Env *ev) : instruction(ev) {}
};

/** system_config for RealScalar               **/
/************************************************/
class system_config_s{
 public:
  int btb_entry;
  int icache_entry;
  int icache_penalty;
  int dcache_entry;
  int dcache_penalty;
  system_config_s(char**);
};

/** evaluation_result for RealScalar           **/
/************************************************/
class evaluation_result_s{
 public:
  /* retired   Conditional branches */
  uint_64t c_branch;     
  uint_64t c_branch_hit;
  uint_64t c_branch_mis;
  /* retired Unconditional branches */
  uint_64t unc_branch;   
  uint_64t bpred_miss;
  uint_64t icache_access;
  uint_64t icache_mis;
  uint_64t dcache_access;
  uint_64t dcache_mis;
  uint_64t clock;
  evaluation_result_s(){
    memset(this, 0, 
	   sizeof(class evaluation_result_s));
  }
};

/************************************************/
class Env_s : public Env{
 public:
  class system_config_s     *sc2;
  class evaluation_result_s *e2;
};

/************************************************/
class chip_wire { /***** Wires *****/
public:
  INST_TYPE ic_out;  /* Instruction Cache Out */

  INST_TYPE dc_wir;  /* Data-Cache Write Ir   */
  data_t    dc_wadr; /* Data-Cache Write Adr  */
  data_t    dc_wdat; /* Data-Cache Write Data */
  INST_TYPE dc_rir;  /* Data-Cache Read  Ir   */
  data_t    dc_radr; /* Data-Cache Read  Adr  */
  data_t    dc_rdat; /* Data-Cache Read  Data */

  int    bpred_tkn;  /* prediction taken/not  */
  data_t bpred_targ; /* prediction target     */
  int    bpred_wtkn; /* taken for update      */
  data_t bpred_wpc;  /*  pc for update        */
  data_t bpred_wnpc; /* npc for update        */

  int icache_stall;  /* stall by I-cache miss */
  int dcache_stall;  /* stall by D-cache miss */
  int   code_stall;  /* stall by code latency */
  int   data_stall;  /* stall by data deps    */

  int flush_pal;     /* 1b: flush by PAL code */
  int flush_bpr;     /* 1b: flush by mis-br   */

  data_t pcgen_pc;   /* PC in PCgen Unit      */
  data_t pcgen_npc;  /* next PC in PCgen      */
  data_t pcgen_bpr;  /* branch miss PC        */
  data_t pcgen_pal;  /* PC by PAL code        */

  instruction_s **stg; /* Pipeline Stage      */
};

/************************************************/
class btb_entry { /*** BTB Entry ***/
 public:
  ADDR_TYPE tag;
  ADDR_TYPE target;
  int predict;
};
/************************************************/
class Bpred{ /*** Branch Target Buffer ***/
  int size;                /* entry size */
  btb_entry *ent;          /* table      */
 public:
  Bpred(Env_s*);           /* Constructor */
  void Init(chip_wire*);   /* Reserved */
  void Step(chip_wire*);   /* Reserved */
  void Update(chip_wire*); /* Reserved */
};

/************************************************/
#define ICLINE_SIZE 64  /* size in byte        */
#define  ICODE_SIZE  4  /* code size in byte   */
#define ICLINE_SFT   6  /* 2^x = ICLINE_SIZE   */
#define  ICODE_SFT   2  /* 2^x = ICODE_SIZE    */
/************************************************/
class icache_line {
 public:
  int valid;
  ADDR_TYPE tag;
  INST_TYPE d[ICLINE_SIZE/ICODE_SIZE]; /* =16 */
};

/************************************************/
class Icache {
  Env_s *ev;
  int size;  /* the number of cache lines */
  int nw;    /* the number of word in line */
  icache_line *buf;
  int stall; /* cache status */
  ADDR_TYPE mis_pc; /* PC of the cache Miss */
 public:
  Icache(Env_s*);
  void ld_line(ADDR_TYPE);
  void Init(chip_wire*);   /* Reserved */
  void Step(chip_wire*);   /* Reserved */
  void Update(chip_wire*); /* Reserved */
};

/************************************************/
#define DCLINE_SIZE 64  /* size in byte, 8 word */
#define  DDATA_SIZE  8  /* data size in byte    */
#define DCLINE_SFT   6  /* 2^x = DCLINE_SIZE    */
#define  DDATA_SFT   3  /* 2^x = DDATA_SIZE     */
/************************************************/
class dcache_line {
public:
  int valid;
  int dirty;
  ADDR_TYPE tag;
  data_t d[DCLINE_SIZE/DDATA_SIZE];
};
/************************************************/
class Dcache {
  Env_s *ev;
  int size;   /* the number of cache lines */
  int nw;     /* the number of word in line */
  dcache_line *buf;
  int stall; /* cache status */
  ADDR_TYPE mis_adr; /* Adr of the cache Miss */
 public:
  Dcache(Env_s*);
  void ld_line(ADDR_TYPE);
  void st_line(ADDR_TYPE);
  void flush();
  void Init(chip_wire*);   /* Reserved */
  void Step(chip_wire*);   /* Reserved */
  void Update(chip_wire*); /* Reserved */
};

/************************************************/
class PCgen{
  Env_s *ev;
 public:
  data_t pc;               /* Program Counter */
  PCgen(Env_s*);           /* Constructor */
  void Init(chip_wire*);   /* Reserved */
  void Step(chip_wire*);   /* Reserved */
  void Update(chip_wire*); /* Reserved */
};

/************************************************/
class InstBuf{ /*** Instruction_S Buffer ***/
  instruction_s **buf;
public:
  InstBuf(Env*);        /* Constructor         */
  instruction_s *alloc(); /* code allocatior   */
};

/************************************************/
class scalar_chip{
  void IDecode();       /* Pipeline stg1       */
  void RegRead();       /* Pipeline stg2       */
  void Execute();       /* Pipeline stg3       */
  void WriteBack();     /* Pipeline stg5       */
  void Init();          /* Reserved            */
  void Step();          /* Reserved            */
  void Update1();       /* Reserved            */
  void Update2();       /* Reserved            */
  void Update3();       /* Reserved            */
  PCgen  *pc_gen;       /* PC Generation U     */
  Bpred  *b_pred;       /* Branch Prediction U */
  Icache *icache;       /* Instruction Cache U */
  Dcache *dcache;       /* Data        Cache U */
  chip_wire *w;         /* set of wires        */
  InstBuf *inst_buf;    /* Instruction Buffer  */
  instruction_s *Inull; /* NULL instruction    */
  int insn_latency;     /* Instruction latency */
public:
  Env_s *ev;            /* Environment         */
  simple_chip *s_chip;  /* Simple Chip         */
  instruction_s **stg;  /* Pipeline Stage      */
  scalar_chip(char*, char**); /* constructor   */
  ~scalar_chip();             /* destructor    */
  int latency(char, int);
  void print_pipeline();
  void house_keeping();
  void verify();
  void main_loop();
  void print_result(); /* print evaluation res */
};
/************************************************/
